@@ -2,24 +2,72 @@ package Wiki::Toolkit::Setup::SQLite;
 
 use strict;
 
-use vars qw( @ISA $VERSION );
+use vars qw( @ISA $VERSION $SCHEMA_VERSION );
 
 use Wiki::Toolkit::Setup::Database;
 
 @ISA = qw( Wiki::Toolkit::Setup::Database );
-$VERSION = '0.09';
+$VERSION = '0.10';
 
 use DBI;
 use Carp;
 
-my %create_sql = (
-    schema_info => "
+$SCHEMA_VERSION = $VERSION*100;
+
+my $create_sql = {
+    8 => {
+        schema_info => [ qq|
 CREATE TABLE schema_info (
   version   integer      NOT NULL default 0
 );
-",
+|, qq|
+INSERT INTO schema_info VALUES (8)
+| ],
+        node => [ qq|
+CREATE TABLE node (
+  id        integer      NOT NULL PRIMARY KEY AUTOINCREMENT,
+  name      varchar(200) NOT NULL DEFAULT '',
+  version   integer      NOT NULL default 0,
+  text      mediumtext   NOT NULL default '',
+  modified  datetime     default NULL
+)
+| ],
+        content => [ qq|
+CREATE TABLE content (
+  node_id   integer      NOT NULL,
+  version   integer      NOT NULL default 0,
+  text      mediumtext   NOT NULL default '',
+  modified  datetime     default NULL,
+  comment   mediumtext   NOT NULL default '',
+  PRIMARY KEY (node_id, version)
+)
+| ],
+        internal_links => [ qq|
+CREATE TABLE internal_links (
+  link_from varchar(200) NOT NULL default '',
+  link_to   varchar(200) NOT NULL default '',
+  PRIMARY KEY (link_from, link_to)
+)
+| ],
+        metadata => [ qq|
+CREATE TABLE metadata (
+  node_id        integer      NOT NULL,
+  version        integer      NOT NULL default 0,
+  metadata_type  varchar(200) NOT NULL DEFAULT '',
+  metadata_value mediumtext   NOT NULL DEFAULT ''
+)
+| ]
+    },
+    9 => {
+        schema_info => [ qq|
+CREATE TABLE schema_info (
+  version   integer      NOT NULL default 0
+);
+|, qq|
+INSERT INTO schema_info VALUES (9)
+| ],
 
-    node => "
+        node => [ qq|
 CREATE TABLE node (
   id        integer      NOT NULL PRIMARY KEY AUTOINCREMENT,
   name      varchar(200) NOT NULL DEFAULT '',
@@ -28,8 +76,8 @@ CREATE TABLE node (
   modified  datetime     default NULL,
   moderate  boolean      NOT NULL default '0'
 )
-",
-    content => "
+| ],
+        content => [ qq|
 CREATE TABLE content (
   node_id   integer      NOT NULL,
   version   integer      NOT NULL default 0,
@@ -39,24 +87,86 @@ CREATE TABLE content (
   moderated boolean      NOT NULL default '1',
   PRIMARY KEY (node_id, version)
 )
-",
-    internal_links => "
+| ],
+        internal_links => [ qq|
 CREATE TABLE internal_links (
   link_from varchar(200) NOT NULL default '',
   link_to   varchar(200) NOT NULL default '',
   PRIMARY KEY (link_from, link_to)
 )
-",
-    metadata => "
+| ],
+        metadata => [ qq|
 CREATE TABLE metadata (
   node_id        integer      NOT NULL,
   version        integer      NOT NULL default 0,
   metadata_type  varchar(200) NOT NULL DEFAULT '',
   metadata_value mediumtext   NOT NULL DEFAULT ''
 )
-"
+| ]
+    },
+    10 => {
+        schema_info => [ qq|
+CREATE TABLE schema_info (
+  version   integer      NOT NULL default 0
+); 
+|, qq|
+INSERT INTO schema_info VALUES (10)
+| ],
+
+        node => [ qq|
+CREATE TABLE node (
+  id        integer      NOT NULL PRIMARY KEY AUTOINCREMENT,
+  name      varchar(200) NOT NULL DEFAULT '',
+  version   integer      NOT NULL default 0,
+  text      mediumtext   NOT NULL default '',
+  modified  datetime     default NULL,
+  moderate  boolean      NOT NULL default '0'
+) 
+|, qq|
+CREATE UNIQUE INDEX node_name ON node (name)
+| ],
+        content => [ qq|
+CREATE TABLE content (
+  node_id   integer      NOT NULL,
+  version   integer      NOT NULL default 0,
+  text      mediumtext   NOT NULL default '',
+  modified  datetime     default NULL,
+  comment   mediumtext   NOT NULL default '',
+  moderated boolean      NOT NULL default '1',
+  verified  datetime     default NULL,
+  verified_info mediumtext   NOT NULL default '',
+  PRIMARY KEY (node_id, version)
+) 
+| ],
+        internal_links => [ qq|
+CREATE TABLE internal_links (
+  link_from varchar(200) NOT NULL default '',
+  link_to   varchar(200) NOT NULL default '',
+  PRIMARY KEY (link_from, link_to)
+) 
+| ],
+        metadata => [ qq|
+CREATE TABLE metadata (
+  node_id        integer      NOT NULL,
+  version        integer      NOT NULL default 0,
+  metadata_type  varchar(200) NOT NULL DEFAULT '',
+  metadata_value mediumtext   NOT NULL DEFAULT ''
+) 
+| ] 
+    },
+};
+
+my %fetch_upgrades = (
+    old_to_8  => 1,
+    old_to_9  => 1,
+    old_to_10 => 1,
+    '8_to_9'  => 1,
+    '8_to_10' => 1,
+    '9_to_10' => 1,
 );
 
+my %upgrades = ();
+
 =head1 NAME
 
 Wiki::Toolkit::Setup::SQLite - Set up tables for a Wiki::Toolkit store in a SQLite database.
@@ -93,52 +203,76 @@ on an existing L<Wiki::Toolkit> database to bring the schema up to date
 with the current L<Wiki::Toolkit> version. If you wish to completely start
 again with a fresh database, run C<cleardb> first.
 
+An optional second argument may be passed specifying the schema version
+to use; this is B<ONLY> intended to be used during unit testing and should
+not normally be specified.
+
 =cut
 
 sub setup {
     my @args = @_;
     my $dbh = _get_dbh( @args );
     my $disconnect_required = _disconnect_required( @args );
+    my $wanted_schema = _get_wanted_schema( @args ) || $SCHEMA_VERSION;
+
+    die "No schema information for requested schema version $wanted_schema\n"
+            unless $create_sql->{$wanted_schema};
 
     # Check whether tables exist, set them up if not.
-    my %tables = fetch_tables_listing($dbh);
+    my %tables = fetch_tables_listing($dbh, $wanted_schema);
 
     # Do we need to upgrade the schema?
     # (Don't check if no tables currently exist)
     my $upgrade_schema;
     my @cur_data; 
     if(scalar keys %tables > 0) {
-        $upgrade_schema = Wiki::Toolkit::Setup::Database::get_database_upgrade_required($dbh,$VERSION);
+        $upgrade_schema = Wiki::Toolkit::Setup::Database::get_database_upgrade_required($dbh,$wanted_schema);
     }
     if($upgrade_schema) {
-        # Grab current data
-        print "Upgrading: $upgrade_schema\n";
-        @cur_data = eval("&Wiki::Toolkit::Setup::Database::fetch_upgrade_".$upgrade_schema."(\$dbh)");
+        if ($fetch_upgrades{$upgrade_schema}) {
+            # Grab current data
+            print "Upgrading: $upgrade_schema\n";
+            @cur_data = eval("&Wiki::Toolkit::Setup::Database::fetch_upgrade_".$upgrade_schema."(\$dbh)");
 
-        # Drop the current tables
-        cleardb($dbh);
+            # Drop the current tables
+            cleardb($dbh);
 
-        # Grab new list of tables
-        %tables = fetch_tables_listing($dbh);
+            # Grab new list of tables
+            %tables = fetch_tables_listing($dbh, $wanted_schema);
+        }
     }
 
     # Set up tables if not found
-    foreach my $required ( keys %create_sql ) {
+    foreach my $required ( keys %{$create_sql->{$wanted_schema}} ) {
         if ( $tables{$required} ) {
             print "Table $required already exists... skipping...\n";
         } else {
             print "Creating table $required... done\n";
-            $dbh->do($create_sql{$required}) or croak $dbh->errstr;
+            foreach my $sql (@{$create_sql->{$wanted_schema}->{$required}} ) {
+                $dbh->do($sql) or croak $dbh->errstr;
+            }
         }
     }
 
-    # Schema version
-    $dbh->do("DELETE FROM schema_info");
-    $dbh->do("INSERT INTO schema_info VALUES (". ($VERSION*100) .")");
-
     # If upgrading, load in the new data
     if($upgrade_schema) {
-        Wiki::Toolkit::Setup::Database::bulk_data_insert($dbh,@cur_data);
+        if ($fetch_upgrades{$upgrade_schema}) {
+            Wiki::Toolkit::Setup::Database::bulk_data_insert($dbh,@cur_data);
+        } else {
+            print "Upgrading schema: $upgrade_schema\n";
+            my @updates = @{$upgrades{$upgrade_schema}};
+            foreach my $update (@updates) {
+                if(ref($update) eq "CODE") {
+                    &$update($dbh);
+                } elsif(ref($update) eq "ARRAY") {
+                    foreach my $nupdate (@$update) {
+                        $dbh->do($nupdate);
+                    }
+                } else {
+                    $dbh->do($update);
+                }
+            } 
+        }
     }
 
     # Clean up if we made our own dbh.
@@ -148,11 +282,12 @@ sub setup {
 # Internal method - what tables are defined?
 sub fetch_tables_listing {
     my $dbh = shift;
+    my $wanted_schema = shift;
 
     # Check whether tables exist, set them up if not.
     my $sql = "SELECT name FROM sqlite_master
                WHERE type='table' AND name in ("
-            . join( ",", map { $dbh->quote($_) } keys %create_sql ) . ")";
+            . join( ",", map { $dbh->quote($_) } keys %{$create_sql->{$wanted_schema}} ) . ")";
     my $sth = $dbh->prepare($sql) or croak $dbh->errstr;
     $sth->execute;
     my %tables;
@@ -194,7 +329,7 @@ sub cleardb {
     print "Dropping tables... ";
     my $sql = "SELECT name FROM sqlite_master
                WHERE type='table' AND name in ("
-            . join( ",", map { $dbh->quote($_) } keys %create_sql ) . ")";
+            . join( ",", map { $dbh->quote($_) } keys %{$create_sql->{$SCHEMA_VERSION}} ) . ")";
     foreach my $tableref (@{$dbh->selectall_arrayref($sql)}) {
         $dbh->do("DROP TABLE $tableref->[0]") or croak $dbh->errstr;
     }
@@ -215,7 +350,7 @@ sub _get_dbh {
         my %args = %{$_[0]};
         if ( $args{dbh} ) {
             return $args{dbh};
-    } else {
+        } else {
             return _make_dbh( %args );
         }
     }
@@ -224,6 +359,19 @@ sub _get_dbh {
     return _make_dbh( dbname => $_[0] );
 }
 
+sub _get_wanted_schema {
+    # Database handle passed in.
+    if ( ref $_[0] and ref $_[0] eq 'DBI::db' ) {
+        return undef;
+    }
+
+    # Args passed as hashref.
+    if ( ref $_[0] and ref $_[0] eq 'HASH' ) {
+        my %args = %{$_[0]};
+        return $args{wanted_schema};
+    }
+}
+
 sub _disconnect_required {
     # Database handle passed in.
     if ( ref $_[0] and ref $_[0] eq 'DBI::db' ) {
@@ -235,7 +383,7 @@ sub _disconnect_required {
         my %args = %{$_[0]};
         if ( $args{dbh} ) {
             return 0;
-    } else {
+        } else {
             return 1;
         }
     }
